from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Iterator,
    List,
    Optional,
    TYPE_CHECKING,
    AsyncIterator,
    Generic,
    TypeVar,
)

from pydantic import (
    AnyHttpUrl,
    Field,
)
from typing_extensions import Self

from async_pixiv.model._base import (
    PixivModel,
    null_dict_validator,
)
from async_pixiv.model.artwork import (
    ArtWork,
    Comment,
    UgoiraMetadata,
)
from async_pixiv.model.novel import (
    Novel,
    NovelMaker,
    NovelSeries,
)
from async_pixiv.model.user import (
    User,
    UserProfile,
    UserProfilePublicity,
    UserWorkSpace,
)

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient

__all__ = [
    'UserPreview', 'UserSearchResult', 'UserIllustsResult',
    'UserBookmarksIllustsResult', 'UserNovelsResult', 'UserDetailResult',
    'UserRelatedResult',

    'IllustSearchResult', 'IllustDetailResult', 'IllustCommentResult',
    'IllustRelatedResult', 'IllustNewResult',

    'UgoiraMetadataResult',

    'RecommendedResult',

    'NovelSearchResult', 'NovelContentResult', 'NovelSeriesResult',
    'NovelDetailResult',
]

T = TypeVar('T')


class PageResult(ABC, PixivModel, Generic[T]):
    next_url: Optional[AnyHttpUrl]
    search_span_limit: Optional[int]

    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        pass

    async def next(
            self, client: Optional["PixivClient"] = None
    ) -> Optional[Self]:
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        if getattr(self, 'next_url', None):
            data = await (await client.get(str(self.next_url))).json()
            return self.__class__.parse_obj(data)
        else:
            return None

    async def iter_all_pages(
            self, client: Optional["PixivClient"] = None
    ) -> AsyncIterator[T]:
        for result in self:
            yield result
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        next_results = await self.next(client)
        while next_results is not None:
            for result in next_results:
                yield result
            next_results = await next_results.next(client)


class UserPreview(PixivModel):
    user: User
    illusts: List[ArtWork]
    is_muted: bool


class UserSearchResult(PageResult[UserPreview]):
    users: List[UserPreview] = Field([], alias='user_previews')

    def __iter__(self) -> Iterator[UserPreview]:
        return iter(self.users)


class UserIllustsResult(PageResult[ArtWork]):
    illusts: List[ArtWork]

    def __iter__(self) -> Iterator[ArtWork]:
        return iter(self.illusts)


class UserBookmarksIllustsResult(UserIllustsResult):
    pass


class UserNovelsResult(PageResult[Novel]):
    novels: List[Novel] = []

    def __iter__(self) -> Iterator[Novel]:
        return iter(self.novels)


class UserDetailResult(PixivModel):
    user: User
    profile: UserProfile
    profile_publicity: UserProfilePublicity
    workspace: UserWorkSpace


class UserRelatedResult(PixivModel):
    users: List[UserPreview] = Field(alias='user_previews')

    def __iter__(self) -> Iterator[UserPreview]:
        return iter(self.users)


class IllustSearchResult(UserIllustsResult):
    pass


class IllustDetailResult(PixivModel):
    illust: ArtWork


class IllustCommentResult(PageResult[Comment]):
    total: int = Field(0, alias='total_comments')
    comments: List[Comment] = []
    next_url: Optional[AnyHttpUrl]
    access_comment: Optional[bool] = Field(alias='comment_access_control')

    def __iter__(self) -> Iterator[Comment]:
        return iter(self.comments)


class IllustRelatedResult(IllustSearchResult):
    pass


class IllustNewResult(IllustDetailResult):
    pass


class UgoiraMetadataResult(PixivModel):
    metadata: UgoiraMetadata = Field(alias="ugoira_metadata")


class RecommendedResult(IllustSearchResult):
    ranking_illusts: List[ArtWork]
    contest_exists: bool


class NovelSearchResult(UserNovelsResult):
    pass


class NovelDetailResult(PixivModel):
    novel: Novel


class NovelContentResult(PixivModel):
    marker: Optional[NovelMaker] = Field(alias='novel_marker')
    content: str = Field(alias='novel_text')
    previous: Optional[Novel] = Field(alias='series_prev')
    next: Optional[Novel] = Field(alias='series_next')

    _check = null_dict_validator('marker', 'previous', 'next')

    @property
    def text(self) -> str:
        return self.content


class NovelSeriesResult(NovelSearchResult):
    series: NovelSeries = Field(alias='novel_series_detail')
    first_novel: Novel = Field(alias='novel_series_first_novel')
    latest_novel: Novel = Field(alias='novel_series_latest_novel')



from datetime import datetime
from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    TYPE_CHECKING,
    Union,
)
from zipfile import ZipFile

from pydantic import (
    AnyHttpUrl,
    Field,
)
from typing_extensions import Literal
from yarl import URL

from async_pixiv.error import ArtWorkTypeError
from async_pixiv.model._base import (
    PixivModel,
    null_dict_validator,
)
from async_pixiv.model.other import (
    ImageUrl,
    Series,
    Tag,
)
from async_pixiv.model.user import User

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient
    from async_pixiv.model.result import (
        IllustDetailResult,
        IllustCommentResult,
        IllustRelatedResult,
    )

__all__ = [
    'ArtWork', 'ArtWorkType',
    'Comment',
    'UgoiraMetadata',
]


class ArtWorkType(Enum):
    illust = 'illust'
    ugoira = 'ugoira'
    manga = 'manga'
    novel = 'novel'

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ArtWorkType):
            return self.value == other.value
        else:
            try:
                return self.value == str(other)
            except (TypeError, ValueError):
                return False


# noinspection PyProtectedMember,PyShadowingBuiltins
class ArtWork(PixivModel):
    class MetaPage(PixivModel):
        image_urls: "ImageUrl"

    class MetaSinglePage(PixivModel):
        original: Optional[AnyHttpUrl] = Field(alias="original_image_url")

    id: int
    title: str
    type: ArtWorkType
    image_urls: "ImageUrl"
    caption: str
    restrict: int
    user: "User"
    tags: List[Tag]
    tools: List[str]
    create_date: datetime
    page_count: int
    width: int
    height: int
    sanity_level: int
    x_restrict: int
    series: Optional[Series]
    meta_single_page: MetaSinglePage
    meta_pages: List[MetaPage]
    total_view: int
    total_bookmarks: int
    is_bookmarked: bool
    visible: bool
    is_muted: bool
    total_comments: Optional[int]
    comment_access_control: Optional[int]

    @property
    def is_nsfw(self) -> bool:
        return self.sanity_level > 5

    @property
    def is_r18(self) -> bool:
        if len(self.tags) >= 1:
            return self.tags[0].name.title().replace('-', '') == 'R18'
        return False

    @property
    def link(self) -> URL:
        return URL(f"https://www.pixiv.net/artworks/{self.id}/")

    @property
    def all_image_urls(self) -> List[URL]:
        if not self.meta_pages:
            # noinspection PyTypeChecker
            return [self.meta_single_page.original]
        result = []
        for page in self.meta_pages:
            result.append(
                URL(
                    str(
                        page.image_urls.original or page.image_urls.large
                    )
                )
            )
        return result

    async def detail(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True
    ) -> "IllustDetailResult":
        from async_pixiv.client._section import SearchFilter

        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()

        return await client.ILLUST.detail(
            self.id,
            filter=SearchFilter.ios if for_ios else SearchFilter.android
        )

    async def comments(
            self, client: Optional["PixivClient"] = None, *,
            offset: Optional[int] = None
    ) -> "IllustCommentResult":
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.ILLUST.comments(self.id, offset=offset)

    async def related(
            self, client: Optional["PixivClient"] = None, *,
            for_ios: bool = True,
            offset: Optional[int] = None,
            seed_id: Optional[int] = None
    ) -> "IllustRelatedResult":
        from async_pixiv.client._section import SearchFilter
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.ILLUST.related(
            self.id, offset=offset, seed_ids=seed_id,
            filter=SearchFilter.ios if for_ios else SearchFilter.android
        )

    async def download(
            self, *,
            full: bool = False,
            output: Optional[Union[str, Path]] = None,
            client: Optional["PixivClient"] = None
    ) -> List[bytes]:
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        if not full or not self.meta_pages:
            return [await client.download(
                str(
                    self.meta_single_page.original or
                    self.image_urls.original or
                    self.image_urls.large
                )
            )]
        else:
            result: List[bytes] = []
            for meta_page in self.meta_pages:
                result.append(
                    await client.download(
                        str(
                            meta_page.image_urls.original or
                            meta_page.image_urls.large
                        ), output=output
                    )
                )
            return result

    async def ugoira_metadata(
            self, client: Optional["PixivClient"] = None
    ) -> Optional["UgoiraMetadata"]:
        if self.type != ArtWorkType.ugoira:
            return None
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return (await client.ILLUST.ugoira_metadata(self.id)).metadata

    async def download_ugoira(
            self, client: Optional["PixivClient"] = None, *,
            type: Literal['zip', 'jpg', 'all', 'iter'] = 'zip'
    ) -> Optional[Union[
        ZipFile, List[bytes], Dict[str, Union[ZipFile, List[bytes]]],
        Iterator[bytes]
    ]]:
        if self.type != ArtWorkType.ugoira:
            raise ArtWorkTypeError(
                'If you want to download a moving image, '
                'please use this method: \"download\"'
            )
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        metadata = await self.ugoira_metadata(client)
        zip_url = metadata.zip_url
        data = await client.download(
            zip_url.original or zip_url.large or zip_url.medium,
        )
        if data is None:
            return None
        zip_file = ZipFile(BytesIO(data))
        if type == 'zip':
            return zip_file
        frames = []
        for frame_info in metadata.frames:
            with zip_file.open(frame_info.file) as f:
                frames.append(f.read())
        if type == 'jpg':
            return frames
        elif type == 'iter':
            return iter(frames)
        else:
            return {'zip': zip_file, 'frames': frames}


class Comment(PixivModel):
    id: int
    comment: str
    date: datetime
    user: User
    parent: Optional["Comment"] = Field(alias='parent_comment')

    _check = null_dict_validator('parent')


class UgoiraMetadata(PixivModel):
    class Frame(PixivModel):
        file: str
        delay: int

    zip_url: ImageUrl = Field(alias='zip_urls')
    frames: List[Frame]



from datetime import datetime
from typing import (
    List,
    Optional,
    TYPE_CHECKING,
)

from pydantic import (
    Extra,
    Field,
)

from async_pixiv.model._base import (
    PixivModel,
    PixivModelConfig,
    null_dict_validator,
)
from async_pixiv.model.other import (
    ImageUrl,
    Series,
    Tag,
)
from async_pixiv.model.user import User

if TYPE_CHECKING:
    from async_pixiv.client import PixivClient
    from async_pixiv.model.result import (
        NovelContentResult,
        NovelSeriesResult,
    )

__all__ = ['Novel', 'NovelMaker', 'NovelSeries']


class Novel(PixivModel):
    id: int
    title: str
    caption: str
    restrict: int
    x_restrict: int
    is_original: bool
    image_url: ImageUrl = Field(alias='image_urls')
    create_date: datetime
    tags: List[Tag]
    page_count: int
    text_length: int
    user: User
    series: Optional[Series]
    is_bookmarked: bool
    total_bookmarks: int
    total_view: int
    visible: bool
    total_comments: int
    is_muted: bool
    is_mypixiv_only: bool
    is_x_restricted: bool
    comment_access_control: Optional[int]

    _check = null_dict_validator('series')

    async def content(
            self, client: Optional["PixivClient"] = None
    ) -> "NovelContentResult":
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.NOVEL.content(self.id)

    async def series_detail(
            self, client: Optional["PixivClient"] = None
    ) -> Optional["NovelSeriesResult"]:
        if self.series is None:
            return None
        if client is None:
            from async_pixiv.client import PixivClient
            client = PixivClient.get_client()
        return await client.NOVEL.series(self.series.id)


class NovelMaker(PixivModel):
    page: int

    class Config(PixivModelConfig):
        extra = Extra.allow


class NovelSeries(Series):
    caption: str
    is_original: bool
    is_concluded: bool
    character_count: int = Field(alias='total_character_count')
    user: User
    display: str = Field(alias='display_text')
    watchlist_added: bool



from typing import (
    Optional,
    TypeVar,
)

from pydantic import (
    BaseConfig as PydanticBaseConfig,
    BaseModel as PydanticBaseModel,
    validator,
)

__all__ = [
    'PixivModel', 'PixivModelConfig',
    'null_dict_validator'
]

T = TypeVar('T')


class PixivModelConfig(PydanticBaseConfig):
    try:
        import ujson as json
    except ImportError:
        import json
    json_dumps = json.dumps
    json_loads = json.loads
    validate_all = True
    validate_assignment = True


class PixivModel(PydanticBaseModel):
    Config = PixivModelConfig

    def __new__(cls, *args, **kwargs):
        cls.update_forward_refs()
        return super(PixivModel, cls).__new__(cls)

    def __str__(self) -> str:
        if hasattr(self, 'id'):
            return f"<{self.__class__.__name__} id=\"{self.id}\">"
        else:
            return f"<{self.__class__.__name__}>"


def null_dict_validator(*fields: str) -> classmethod:
    def v(cls, value: T) -> Optional[T]:
        if value == {}:
            return None
        else:
            return value

    return validator(*fields, pre=True, always=True, allow_reuse=True)(v)



from abc import ABC
from datetime import date
from enum import Enum as BaseEnum
from io import BytesIO
from pathlib import Path
from typing import (
    Dict,
    Iterator,
    List,
    Optional,
    TYPE_CHECKING,
    TypeVar,
    Union,
)
from zipfile import ZipFile

from typing_extensions import Literal
from yarl import URL

from async_pixiv.model.result import (
    IllustCommentResult,
    IllustDetailResult,
    IllustNewResult,
    IllustRelatedResult,
    IllustSearchResult,
    NovelContentResult,
    NovelDetailResult,
    NovelSearchResult,
    NovelSeriesResult,
    RecommendedResult,
    UgoiraMetadataResult,
    UserBookmarksIllustsResult,
    UserDetailResult,
    UserIllustsResult,
    UserNovelsResult,
    UserRelatedResult,
    UserSearchResult,
)
from ..error import ArtWorkTypeError
from ..model.artwork import ArtWorkType

if TYPE_CHECKING:
    from ._client import PixivClient

__all__ = [
    'SearchTarget', 'SearchShort', 'SearchDuration', 'SearchFilter',
    'SectionType',
    'USER', 'ILLUST', 'NOVEL'
]

API_HOST = URL("https://app-api.pixiv.net")
V1_API = API_HOST / 'v1'
V2_API = API_HOST / 'v2'


class Enum(BaseEnum):
    def __str__(self) -> str:
        # noinspection PyTypeChecker
        return self.value


class SearchTarget(Enum):
    partial = 'partial_match_for_tags'  # 标签部分一致
    full = 'exact_match_for_tags'  # 标签完全一致
    text = 'text'  # 正文
    keyword = 'keyword'  # 关键词


class SearchShort(Enum):
    date_desc = 'date_desc'
    date_asc = 'date_asc'
    popular_desc = 'popular_desc'
    popular_asc = 'popular_asc'


class SearchDuration(Enum):
    day = 'within_last_day'
    week = 'within_last_week'
    month = 'within_last_month'


class SearchFilter(Enum):
    android = 'for_android'
    ios = 'for_ios'


# noinspection PyShadowingBuiltins
class _Section(ABC):
    _client: "PixivClient"
    _type: str = ''

    @property
    def type(self) -> str:
        return self._type

    def __init__(self, client: "PixivClient") -> None:
        self._client = client

    async def search(
            self,
            word: str, *,
            sort: Union[
                Literal['date_desc', 'date_asc', 'popular_desc', 'popular_asc'],
                SearchShort
            ] = SearchShort.date_desc,
            target: Union[
                SearchTarget,
                Literal[
                    'partial_match_for_tags', 'keyword',
                    'exact_match_for_tags', 'text'
                ]
            ] = None,
            duration: Optional[Union[
                Literal[
                    'within_last_day',
                    'within_last_week',
                    'within_last_month',
                    'within_last_year'
                ], SearchDuration
            ]] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None, **kwargs
    ) -> Dict:
        request = await self._client.get(
            V1_API / f'search/{self._type}',
            params={
                'word': word, 'sort': sort, 'duration': duration,
                'filter': filter, 'offset': offset, 'search_target': target,
                **kwargs
            }
        )
        return await request.json()

    async def detail(
            self, id: int, *, filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> Dict:
        request = await self._client.get(
            V1_API / f'{self._type}/detail',
            params={f'{self._type}_id': id, 'filter': filter}
        )
        return await request.json()


SectionType = TypeVar('SectionType', bound=_Section)


class IllustType(Enum):
    illust = 'illust'
    manga = 'manga'
    ugoira = 'ugoira'
    novel = 'novel'


# noinspection PyShadowingBuiltins
class USER(_Section):
    _type = 'user'

    async def search(
            self,
            word: str, *,
            sort: Union[
                Literal['date_desc', 'date_asc', 'popular_desc', 'popular_asc'],
                SearchShort
            ] = SearchShort.date_desc,
            duration: Optional[Union[
                Literal[
                    'within_last_day',
                    'within_last_week',
                    'within_last_month',
                    'within_last_year'
                ], SearchDuration
            ]] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None, **kwargs
    ) -> UserSearchResult:
        data = await super(USER, self).search(
            word=word, sort=sort, duration=duration, filter=filter,
            offset=offset, **kwargs
        )
        return UserSearchResult.parse_obj(data)

    async def detail(
            self, id: int, *, filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> UserDetailResult:
        if id is None:
            id = self._client.account.id
        data = await super(USER, self).detail(id=id, filter=filter)
        return UserDetailResult.parse_obj(data)

    async def illusts(
            self, id: Optional[int] = None, *,
            type: Union[
                Literal['illust', 'manga', 'novel', 'ugoira'], IllustType
            ] = IllustType.illust,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None
    ) -> UserIllustsResult:
        if id is None:
            id = self._client.account.id
        data = await (await self._client.get(
            V1_API / "user/illusts",
            params={
                'user_id': id, 'type': type, 'filter': filter, 'offset': offset
            }
        )).json()
        return UserIllustsResult.parse_obj(data)

    async def bookmarks(
            self, id: Optional[int] = None, *,
            tag: Optional[str] = None,
            max_bookmark_id: Optional[int] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
    ) -> UserBookmarksIllustsResult:
        if id is None:
            id = self._client.account.id
        data = await (await self._client.get(
            V1_API / "user/bookmarks/illust",
            params={
                'user_id': id, 'filter': filter, 'tag': tag,
                'max_bookmark_id': max_bookmark_id, 'restrict': 'public'
            }
        )).json()
        return UserBookmarksIllustsResult.parse_obj(data)

    async def novels(
            self, id: Optional[int] = None, *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: int = None
    ) -> UserNovelsResult:
        id = self._client.account.id if id is None else id
        data = await (await self._client.get(
            V1_API / "user/novels",
            params={'user_id': id, 'filter': filter, 'offset': offset}
        )).json()
        return UserNovelsResult.parse_obj(data)

    async def related(
            self, id: Optional[int] = None, *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: int = None,
    ) -> UserRelatedResult:
        if id is None:
            id = self._client.account.id
        data = await (await self._client.get(
            V1_API / "user/related",
            params={
                'seed_user_id': id, 'filter': filter, 'offset': offset
            }
        )).json()
        return UserRelatedResult.parse_obj(data)


# noinspection PyShadowingBuiltins
class ILLUST(_Section):
    _type = 'illust'

    async def follow(
            self, *, offset: Optional[int] = None
    ) -> IllustSearchResult:
        data = await (await self._client.get(
            V2_API / "illust/follow",
            params={'restrict': 'public', 'offset': offset}
        )).json()
        return IllustSearchResult.parse_obj(data)

    async def search(
            self,
            word: str, *,
            sort: Union[
                Literal['date_desc', 'date_asc', 'popular_desc', 'popular_asc'],
                SearchShort
            ] = SearchShort.date_desc,
            target: Union[
                SearchTarget,
                Literal[
                    'partial_match_for_tags', 'keyword',
                    'exact_match_for_tags', 'text'
                ]
            ] = SearchTarget.partial,
            duration: Optional[Union[
                Literal[
                    'within_last_day',
                    'within_last_week',
                    'within_last_month',
                ], SearchDuration
            ]] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None,
            min_bookmarks: Optional[int] = None,
            max_bookmarks: Optional[int] = None,
            start_date: Optional[date] = None,
            end_date: Optional[date] = None,
            **kwargs
    ) -> IllustSearchResult:
        if start_date is not None:
            start_date = start_date.strftime('%Y-%m-%d')
        if end_date is not None:
            end_date = end_date.strftime('%Y-%m-%d')
        data = await super(ILLUST, self).search(
            word=word, sort=sort, duration=duration, filter=filter,
            offset=offset, bookmark_num_min=min_bookmarks,
            bookmark_num_max=max_bookmarks, target=target,
            start_date=start_date, end_date=end_date,
        )
        return IllustSearchResult.parse_obj(data)

    async def detail(
            self, id: int, *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> IllustDetailResult:
        data = await super(ILLUST, self).detail(id, filter=filter)
        return IllustDetailResult.parse_obj(data)

    async def comments(
            self, id: int, *, offset: Optional[int] = None
    ) -> IllustCommentResult:
        data = await (await self._client.get(
            V1_API / "illust/comments", params={
                'illust_id': id, "offset": offset
            }
        )).json()
        return IllustCommentResult.parse_obj(data)

    async def related(
            self, id: int, *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None,
            seed_ids: Optional[Union[List[int], int]] = None
    ) -> IllustRelatedResult:
        data = await (await self._client.get(
            V2_API / "illust/related", params={
                'illust_id': id, "offset": offset, 'filter': filter,
                'seed_illust_ids': (
                    [seed_ids] if seed_ids is not list else seed_ids
                )
            }
        )).json()
        return IllustRelatedResult.parse_obj(data)

    async def recommended(
            self, *, with_auth: bool = True, type: Union[
                Literal['illust', 'manga', 'novel', 'ugoira'], IllustType
            ] = IllustType.illust,
            include_ranking_label: bool = True,
            include_ranking_illusts: Optional[bool] = None,
            offset: Optional[int] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            max_bookmark_id_for_recommend: Optional[int] = None,
            min_bookmark_id_for_recent_illust: Optional[int] = None,
            bookmark_illust_ids: Optional[Union[List[int], int]] = None,
            include_privacy_policy: Optional[Union[
                str, List[Union[int, str]]
            ]] = None,
            viewed: Optional[List[int]] = None
    ) -> RecommendedResult:
        data = await (await self._client.get(
            V1_API / (
                "illust/recommended"
                if with_auth else
                "illust/recommended-nologin"
            ), params={
                'content_type': type,
                'include_ranking_label': include_ranking_label,
                'include_ranking_illusts': include_ranking_illusts,
                'offset': offset, 'filter': filter, 'viewed': viewed,
                'max_bookmark_id_for_recommend': max_bookmark_id_for_recommend,
                'min_bookmark_id_for_recent_illust':
                    min_bookmark_id_for_recent_illust,
                'bookmark_illust_ids': bookmark_illust_ids,
                'include_privacy_policy': include_privacy_policy,
            }
        )).json()
        return RecommendedResult.parse_obj(data)

    async def new_illusts(
            self, type: Literal['illust', 'manga'] = 'illust', *,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            max_illust_id: Optional[int] = None
    ) -> IllustNewResult:
        data = await (await self._client.get(
            V1_API / 'illust/new',
            params={
                'content_type': type,
                'max_illust_id': max_illust_id,
                'filter': filter,
            }
        )).json()
        return IllustNewResult.parse_obj(data)

    async def ugoira_metadata(self, id: int) -> UgoiraMetadataResult:
        data = await (await self._client.get(
            V1_API / 'ugoira/metadata', params={'illust_id': id}
        )).json()
        return UgoiraMetadataResult.parse_obj(data)

    async def download(
            self, id: int, *,
            full: bool = False,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            output: Optional[Union[str, Path, BytesIO]] = None
    ) -> List[bytes]:
        artwork = (await self.detail(id, filter=filter)).illust
        if artwork.type == ArtWorkType.ugoira:
            raise ArtWorkTypeError(
                'If you want to download a moving image, '
                'please use this method: \"download_ugoira\"'
            )
        if not full or not artwork.meta_pages:
            return [await self._client.download(
                str(
                    artwork.meta_single_page.original or
                    artwork.image_urls.original or
                    artwork.image_urls.large
                ), output=output
            )]
        else:
            result: List[bytes] = []
            for meta_page in artwork.meta_pages:
                result.append(
                    await self._client.download(
                        str(
                            meta_page.image_urls.original or
                            meta_page.image_urls.large
                        ), output=output
                    )
                )
            return result

    async def download_ugoira(
            self, id: int, *,
            type: Literal['zip', 'jpg', 'all', 'iter'] = 'zip',
    ) -> Optional[Union[
        ZipFile, List[bytes], Dict[str, Union[ZipFile, List[bytes]]],
        Iterator[bytes]
    ]]:
        metadata = (await self.ugoira_metadata(id)).metadata
        zip_url = metadata.zip_url
        data = await self._client.download(
            zip_url.original or zip_url.large or zip_url.medium,
        )
        if data is None:
            return None
        zip_file = ZipFile(BytesIO(data))
        if type == 'zip':
            return zip_file
        frames = []
        for frame_info in metadata.frames:
            with zip_file.open(frame_info.file) as f:
                frames.append(f.read())
        if type == 'jpg':
            return frames
        elif type == 'iter':
            return iter(frames)
        else:
            return {'zip': zip_file, 'frames': frames}


# noinspection PyShadowingBuiltins
class NOVEL(_Section):
    _type = "novel"

    async def search(
            self,
            word: str, *,
            sort: Union[
                Literal['date_desc', 'date_asc', 'popular_desc', 'popular_asc'],
                SearchShort
            ] = SearchShort.date_desc,
            target: Union[
                SearchTarget,
                Literal[
                    'partial_match_for_tags', 'keyword',
                    'exact_match_for_tags', 'text'
                ]
            ] = SearchTarget.partial,
            duration: Optional[Union[
                Literal[
                    'within_last_day',
                    'within_last_week',
                    'within_last_month',
                    'within_last_year'
                ], SearchDuration
            ]] = None,
            filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios,
            offset: Optional[int] = None,
            start_date: Optional[date] = None,
            end_date: Optional[date] = None,
            merge_plain_keyword_results: bool = True,
            include_translated_tag_results: bool = True, **kwargs
    ) -> NovelSearchResult:
        if start_date is not None:
            start_date = start_date.strftime('%Y-%m-%d')
        if end_date is not None:
            end_date = end_date.strftime('%Y-%m-%d')
        data = await super(NOVEL, self).search(
            word, sort=sort, target=target, duration=duration, filter=filter,
            offset=offset, start_date=start_date, end_date=end_date,
            merge_plain_keyword_results=merge_plain_keyword_results,
            include_translated_tag_results=include_translated_tag_results
        )
        return NovelSearchResult.parse_obj(data)

    async def detail(
            self, id: int, *, filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> NovelDetailResult:
        data = await (await self._client.get(
            V2_API / 'novel/detail', params={'novel_id': id}
        )).json()
        return NovelDetailResult.parse_obj(data)

    async def content(self, id: int) -> NovelContentResult:
        data = await (await self._client.get(
            V1_API / 'novel/text', params={'novel_id': id}
        )).json()
        return NovelContentResult.parse_obj(data)

    async def series(
            self, id: int, *, filter: Optional[Union[
                Literal['for_android', 'for_ios'], SearchFilter
            ]] = SearchFilter.ios
    ) -> NovelSeriesResult:
        data = await (await self._client.get(
            V2_API / 'novel/series', params={'series_id': id, 'filter': filter}
        )).json()
        return NovelSeriesResult.parse_obj(data)



from typing import Optional

from pydantic import HttpUrl

from async_pixiv.model._base import PixivModel

__all__ = [
    'Series',
    'ImageUrl',
    'Tag', 'TagTranslation',
]


class Series(PixivModel):
    id: int
    title: str


class ImageUrl(PixivModel):
    square_medium: Optional[HttpUrl]
    medium: Optional[HttpUrl]
    large: Optional[HttpUrl]
    original: Optional[HttpUrl]


class TagTranslation(PixivModel):
    zh: Optional[str]
    en: Optional[str]
    jp: Optional[str]


class Tag(PixivModel):
    name: str
    translated_name: Optional[str]
    added_by_uploaded_user: Optional[bool]
